/*
 * main.c - main file for GG's Blocks
 *      - all the application code and drawing code is in here
 *
 *    Copyright (C) 2003 Eric Cheung.
 *
 *    This file is part of GG's Blocks.
 *
 *    GG's Blocks is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    GG's Blocks is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with GG's Blocks; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <StringMgr.h>
#include <TimeMgr.h>
#include <DataMgr.h>
#include <Font.h>
#include "main.h"
#include "ggblocks.h"

// Constants
// dimensions of the playing field. Fixed to 160x160
// top left corner of playfield
#define XSTART      3
#define YSTART      17
// width and height of playfield
#define XLEN        154
#define YLEN        126
// dimensions of a single square
#define XSQ         11
#define YSQ         14

// Global variables
// these six form the gamestate which is saved on exit
WORLD world;
Int32 score;
Int16 level;
Int32 highScore[3];
Int16 highLevel[3];
Boolean gameOver;       // true if it's game over

// other globals
static DmOpenRef db;
static Boolean isColor;       // whether we're running on a colour palm or not
static UInt16 ticksPerSec;
static RGBColorType blackRGB = {0,0,0,0};
static IndexedColorType blackIndex;
static Boolean doFullRedraw;

// Function prototypes
UInt32  PilotMain(UInt16, MemPtr, UInt16);

static Boolean mainFormHandler(EventType *);
static void initApplication(void);
static Boolean applicationHandler(EventType *);
static void eventLoop(void);
static void endApplication(void);

// resets the high scores to initial values
static void resetHighScores(void);
// draws the gameboard
static void displayWorld(WORLD *w);
// draws a single tile
static void drawTile(RectangleType *rect, TILE t);
// put the level description string in lvlStr.
static void levelLabel(Char *lvlStr);
// create a new game
static void newGame(void);
// handle saving/restoring state
// returns 1 on success
static Boolean startDatabase(void);
static void writeDatabase(void);
// the database is implicitly written out when stop is called
static void stopDatabase(void);
// writes and reads records from the global database db
static void newRecord(UInt16 index, UInt32 size);
static void readRecord(UInt16 index, UInt32 size, MemPtr ptr);
static void writeRecord(UInt16 index, UInt32 size, MemPtr ptr);

// MainForm event handler
static Boolean mainFormHandler(EventType *event) {
    Boolean processed = false;
    EventType newEvent;
    Char s[32];
    Char tmpStr[maxStrIToALen];
    Int16 r, c, i;
    Int32 moveScore;
    UInt16 mainFormLabelIndex;
    UInt16 mainFormScoreIndex;
    FormType *fP;
   
    // get references to the main form, and the two labels inside it
    fP = FrmGetActiveForm();
    mainFormLabelIndex = FrmGetObjectIndex(fP, MainFormLabel);
    mainFormScoreIndex = FrmGetObjectIndex(fP, MainFormScore);

    switch (event->eType) {
    case frmOpenEvent:
        FrmDrawForm(fP);

        // create and set the bottom label
        levelLabel(s);
        FrmHideObject(fP, mainFormLabelIndex);
        FrmCopyLabel(fP, MainFormLabel, s);
        FrmShowObject(fP, mainFormLabelIndex);

        // mark the whole field for redraw
        doFullRedraw = true;

        // send an update event to ourselves so we redraw the field
        MemSet(&newEvent, sizeof(EventType), 0);
        newEvent.eType = frmUpdateEvent;
        newEvent.data.frmUpdate.formID = FrmGetActiveFormID();
        EvtAddEventToQueue(&newEvent);

        processed = true;
        break;

    case frmUpdateEvent:
        // set the score
        StrIToA(s, score);
        FrmHideObject(fP, mainFormScoreIndex);
        FrmCopyLabel(fP, MainFormScore, s);
        FrmShowObject(fP, mainFormScoreIndex);

        // draw seperators
        if(isColor)
            WinSetForeColor(blackIndex);

        if(doFullRedraw){
            doFullRedraw = false;
            FrmDrawForm(fP);
            worldMarkModified(&world);
        }

        WinDrawLine(0,145,159,145);
        WinDrawLine(0,146,159,146);

        // draw the world
        displayWorld(&world);

        if(!world.movesLeft && !gameOver){
            Int32 bonus = worldFinalPointValue(&world);

            // we've completed a level...
            Int32 threshold = (Int32)level*level*750;

            score+=bonus;

            // show the score with bonus added on
            StrIToA(s, score);
            FrmHideObject(fP, mainFormScoreIndex);
            FrmCopyLabel(fP, MainFormScore, s);
            FrmShowObject(fP, mainFormScoreIndex);

            // do we qualify for the next round?
            if(score >= threshold){
                // new level
                worldGenerate(&world);

                // tell the user they've passed this level
                StrIToA(s, level);
                StrIToA(tmpStr, bonus);
                doFullRedraw = true;    // need to redraw behind Alert box
                FrmCustomAlert(MainFormNextLvlAlert, s, tmpStr, NULL);

                level++;
                
                // redraw the playfield
                // call to worldMarkModified seems redundant. it is here because
                // for some reason, on debug roms whenever FrmXXXAlert is
                // called an implicit update event is sent! which screws us over
                // so we need this
                worldMarkModified(&world);
                displayWorld(&world);
            }
            else{
                // game over
                gameOver = true;

                // have we set a high score? i represents the insertion point
                for(i=3;highScore[i-1]<=score && i>0;i--);

                if(i<3) {
                    // yes; write it into the correct place and move records
                    // down
                    MemMove(&highScore[i+1],&highScore[i],(2-i)*sizeof(Int32));
                    MemMove(&highLevel[i+1],&highLevel[i],(2-i)*sizeof(Int16));
                    highScore[i] = score;
                    highLevel[i] = level;
                    doFullRedraw = true;    // need to redraw behind Alert box
                    FrmAlert(NewHighScoreAlert);

                    // make the high score screen pop up by generating
                    // a menu click on high scores
                    MemSet(&newEvent, sizeof(EventType), 0);
                    newEvent.eType = menuEvent;
                    newEvent.data.menu.itemID = MainGameHighScore;
                    EvtAddEventToQueue(&newEvent);
                }
            }

            // set the bottom label to reflect whether we're game over or
            // a new level
            levelLabel(s);
            FrmHideObject(fP, mainFormLabelIndex);
            FrmCopyLabel(fP, MainFormLabel, s);
            FrmShowObject(fP, mainFormLabelIndex);
        }

        processed = true;

        break;

    case penDownEvent:
        // don't catch clicks outside the playfield
        if(event->screenX < XSTART || event->screenX >= XSTART+XLEN ||
                event->screenY < YSTART || event->screenY >= YSTART+YLEN)
            break;

        // if the game is already finished, treat a pen press as a request for
        // a new game
        if(gameOver) {
            newGame();
            break;
        }

        c = (event->screenX-XSTART)*WORLDCOL/XLEN;
        r = (event->screenY-YSTART)*WORLDROW/YLEN;

        // make the move
        moveScore = worldApplyMove(&world, c, r);

        // was this a valid move?
        if(moveScore!=-1){
            // show the board with EMPTYs...
            displayWorld(&world);
                
            // wait
            if(ticksPerSec != 0)
                SysTaskDelay(MOVEDELAY*ticksPerSec/100);
            
            score += moveScore;

            // take out the empty spaces
            worldCompact(&world);

            // send an update event to ourselves so we redraw the field
            MemSet(&newEvent, sizeof(EventType), 0);
            newEvent.eType = frmUpdateEvent;
            newEvent.data.frmUpdate.formID = FrmGetActiveFormID();
            EvtAddEventToQueue(&newEvent);
        }

        processed = true;
        break;

    default:
        break;
    }

    return processed;
}

// application initialization function
void initApplication(void) {
    UInt32 romVer;

    // do we support color?
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVer);

    // only palm OS 3.5 and above support color at all...
    if(romVer >= sysMakeROMVersion(3,5,0,sysROMStageRelease,0))
        WinScreenMode(winScreenModeGetSupportsColor,NULL,NULL,NULL,&isColor);
    else
        isColor = false;
        
    // only palm OS 2.0 and above support proper tick counting stuff
    // if it's OS 1.0, set 0 to denote don't delay at all, because OS 1.0
    // scares me as it's so buggy
    ticksPerSec = romVer>=sysMakeROMVersion(2,0,0,sysROMStageRelease,0)?
            SysTicksPerSecond():0;

    if(isColor)
        // determine index of black
        blackIndex = WinRGBToIndex(&blackRGB);

    doFullRedraw = false;

    // read in state variables
    startDatabase();

    // goto the main form
    FrmGotoForm(MainForm);

}

// PilotMain entry function
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags) {
    if (cmd == sysAppLaunchCmdNormalLaunch) 
    {
        // run the application
        initApplication();
        eventLoop();
        endApplication();
    }

    return 0;
}

// application event handling
Boolean applicationHandler(EventType *event)
{
    FormType *frm;
    UInt16 i;
    Char highScoreStr[3][32];
    Char tmp[maxStrIToALen];
    Boolean processed = false;

    switch (event->eType) {
    case frmLoadEvent:
        {
            UInt16   formID = event->data.frmLoad.formID;
            frm   = FrmInitForm(formID);

            FrmSetActiveForm(frm);
            switch (formID) 
            {
            case MainForm:
                FrmSetEventHandler(frm, 
                    (FormEventHandlerPtr)mainFormHandler);

                processed = true;
                break;

            default:
                break;
            }
        }
        break;
         
    case menuEvent:
        switch(event->data.menu.itemID) {
        case MainGameNewGame:
            newGame();
            break;

        case MainGameHighScore:
            for(i=0;i<3;i++){
                StrIToA(highScoreStr[i], highLevel[i]);
                StrCat(highScoreStr[i], "          ");
                StrIToA(tmp, highScore[i]);
                StrCat(highScoreStr[i], tmp);
            }

            doFullRedraw = true;    // need to redraw behind Alert box
            if(FrmCustomAlert(DisplayHighScoreAlert, highScoreStr[0], 
                                highScoreStr[1], highScoreStr[2])==1){
                // user's pressed reset scores button
                doFullRedraw = true;    // need to redraw behind Alert box
                if(FrmAlert(ResetConfirmationAlert)==0)
                    // yes, reset the high scores
                    resetHighScores();
            }
            break;

        case MainGameTheBasics:
            doFullRedraw = true;    // need to redraw behind Alert box
            FrmAlert(TheBasicsAlert);
            doFullRedraw = true;    // need to redraw behind Alert box
            FrmAlert(TheBasics2Alert);
            doFullRedraw = true;    // need to redraw behind Alert box
            FrmAlert(TheBasics3Alert);
            doFullRedraw = true;    // need to redraw behind Alert box
            FrmAlert(TheBasics4Alert);
            break;

        case MainGameAbout:
            doFullRedraw = true;    // need to redraw behind Alert box
            FrmAlert(AboutAlert);
            break;

        default:
            break;
        }
        break;

    default:
         break;
    }

    return processed;
}

// application event processing loop
void eventLoop(void) {
    EventType event;
    FormType  *frm;
    UInt16    err;

    do {
        EvtGetEvent(&event, evtWaitForever);
        frm = FrmGetActiveForm();

        if (!SysHandleEvent(&event)) 
            if (!MenuHandleEvent(0, &event, &err)) 
                if (!applicationHandler(&event)) 
                    FrmDispatchEvent(&event);

    } while(event.eType != appStopEvent);
}

// cleanup
void endApplication(void) {
    FrmCloseAllForms();
    // this will implicitly write out the application state
    stopDatabase();
}

// draws a rectangle in the approriate pattern
static void drawTile(RectangleType *rect, TILE t) {
    static const CustomPatternType white = {0,0,0,0,0,0,0,0};
    static const RGBColorType whiteRGB = {0,255,255,255};
    static IndexedColorType whiteIndex;
    static const CustomPatternType mixer =
        {0xFF,0x00,0x00,0x00,0xFF,0x00,0x00,0x00};
    static const RGBColorType mixerRGB = {0,255,0,255};
    static IndexedColorType mixerIndex;
    // note - the 0th pattern is the pattern used by all squares when
    // colour is enabled
    static const CustomPatternType tilePattern[NUMDIFFTILES] =
        {   {0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF},
            {0xAA,0x55,0xAA,0x55,0xAA,0x55,0xAA,0x55},
            {0xF0,0xF0,0xF0,0xF0,0x0F,0x0F,0x0F,0x0F},
            {0xFF,0x99,0x99,0xFF,0xFF,0x99,0x99,0xFF},
            {0x11,0x11,0x11,0x11,0x11,0x11,0x11,0x11}   };
    static const RGBColorType tileRGB[NUMDIFFTILES] =
        {   {0,0,0,0},
            {0,255,0,0},
            {0,0,255,0},
            {0,0,0,255},
            {0,255,255,0}   };
    static IndexedColorType tilePatternIndex[NUMDIFFTILES];
    static Boolean firstTime = true;
    Coord centerX;

    if(firstTime){
        UInt16 i;

        if(isColor){
            // determine the closest index to each color
            whiteIndex = WinRGBToIndex(&whiteRGB);
            mixerIndex = WinRGBToIndex(&mixerRGB);
            for(i=0;i<NUMDIFFTILES;i++)
                tilePatternIndex[i] = WinRGBToIndex(&tileRGB[i]);
        }

        firstTime = false;
    }

    switch(t){
    case EMPTY:
        // clear this area
        if(isColor)
            WinSetForeColor(whiteIndex);
        WinSetPattern(&white);
        WinFillRectangle(rect, 0);
        break;

    case MIXER:
        if(isColor)
            WinSetForeColor(mixerIndex);
        WinSetPattern(&mixer);
        WinFillRectangle(rect, 0);
        FntSetFont(1);
        centerX = (rect->topLeft.x+(rect->extent.x/2))-((FntCharWidth('?'))/2);
        WinDrawChars("?", 1, centerX, rect->topLeft.y+1);
        break;

    default:
        if(isColor){
            WinSetForeColor(tilePatternIndex[t]);
            WinSetPattern(&tilePattern[0]);
        }
        else
            WinSetPattern(&tilePattern[t]);
        WinFillRectangle(rect, 0);
        break;

    }
}

// draws the gameboard
// note we only display the regions marked as modified. see comment in 
// ggblocks.c to find out how this works.
static void displayWorld(WORLD *w){
    Int16 col, row;
    RectangleType rect;

    rect.extent.x = XSQ-1;
    rect.extent.y = YSQ-1;
    rect.topLeft.x = XSTART;

    for(col=0;col<WORLDCOL;col++,rect.topLeft.x+=XSQ){
        rect.topLeft.y = YSTART+YSQ*w->topUpdate[col];

        for(row=w->topUpdate[col];row<=w->bottomUpdate[col];row++,
                        rect.topLeft.y+=YSQ)
            drawTile(&rect, w->tile[col][row]);

        // mark this range as 'don't need to be redrawn'
        w->topUpdate[col] = WORLDROW-1;
        w->bottomUpdate[col] = 0;
    }
}

// make the level label
static void levelLabel(Char *lvlStr) {
    Char tmpStr[maxStrIToALen];

    StrCopy(lvlStr, "Lvl ");
    StrIToA(tmpStr, level);
    StrCat(lvlStr, tmpStr);

    if(!gameOver){
        // form the bottom label to say what the next level
        // threshold is
        StrCat(lvlStr, " - Next Lvl ");
        StrIToA(tmpStr, (Int32)level*level*750);
        StrCat(lvlStr, tmpStr);
    }
    else{
        // tell the user it's game over
        StrCat(lvlStr, " - Game Over");
    }
}

// create a new game
static void newGame(void){
    Char s[32];
    EventType newEvent;
    FormType *fP;
    UInt16 mainFormLabelIndex;

    gameOver = false;

    // set initial values
    score = 0;
    level = 1;

    // reset the board
    worldGenerate(&world);
    // and create the level label
    levelLabel(s);

    fP = FrmGetFormPtr(MainForm); 
    mainFormLabelIndex = FrmGetObjectIndex(fP, MainFormLabel);

    // redraw the bottom label
    FrmHideObject(fP, mainFormLabelIndex);
    FrmCopyLabel(fP, MainFormLabel, s);
    FrmShowObject(fP, mainFormLabelIndex);
       
    doFullRedraw = true;

    // send an update event to mainform
    MemSet(&newEvent, sizeof(EventType), 0);
    newEvent.eType = frmUpdateEvent;
    newEvent.data.frmUpdate.formID = MainForm;
    EvtAddEventToQueue(&newEvent);
}

static void resetHighScores(void) {
        highScore[0] = highScore[1] = highScore[2] = INITIALHIGHSCORE;
        highLevel[0] = highLevel[1] = highLevel[2] = INITIALHIGHLEVEL;
}

// returns 1 on success
static Boolean startDatabase(void)
{
    UInt16 index = 0;

    db = DmOpenDatabaseByTypeCreator(DBID, CREATORID, dmModeReadWrite);

    if (!db) {
        // doesn't exist - create a new one
        if (DmCreateDatabase(0, DBNAME, CREATORID, DBID, false))
            return 0;

        db = DmOpenDatabaseByTypeCreator(DBID, CREATORID, dmModeReadWrite);

        // create starting values
        worldGenerate(&world);
        gameOver = false;
        score = 0;
        level = 1;
        resetHighScores();

        // create new records for
        // world, score, level, highScore, highLevel, gameOver
        newRecord(index++, sizeof(WORLD));
        newRecord(index++, sizeof(Int32));
        newRecord(index++, sizeof(Int16));
        newRecord(index++, sizeof(Int32)*3);
        newRecord(index++, sizeof(Int16)*3);
        newRecord(index++, sizeof(Boolean));

        // store starting values in newly created records
        writeDatabase();
    }
    else {
        // read in the state variables from the database
        // world, score, level, highScore, highLevel, gameOver
        readRecord(index++,sizeof(WORLD), &world);
        readRecord(index++,sizeof(Int32), &score);
        readRecord(index++,sizeof(Int16), &level);
        readRecord(index++,sizeof(Int32)*3, highScore);
        readRecord(index++,sizeof(Int16)*3, highLevel);
        readRecord(index++,sizeof(Boolean), &gameOver);
    }

    return 1;
}

static void writeDatabase(void) {
    UInt16 index = 0;

    // write state variables in, each in its own record
    // world, score, level, highScore, highLevel, gameOver
    writeRecord(index++, sizeof(WORLD), &world);
    writeRecord(index++, sizeof(Int32), &score);
    writeRecord(index++, sizeof(Int16), &level);
    writeRecord(index++, sizeof(Int32)*3, highScore);
    writeRecord(index++, sizeof(Int16)*3, highLevel);
    writeRecord(index++, sizeof(Boolean), &gameOver);
}

static void stopDatabase(void) {
    writeDatabase();
    DmCloseDatabase(db);
}

static void newRecord(UInt16 index, UInt32 size){
    DmNewRecord(db, &index, size);
    DmReleaseRecord(db, index, true);  
}

static void writeRecord(UInt16 index, UInt32 size, MemPtr ptr){
    MemPtr recPtr;
    MemHandle recHandle;

    recHandle = DmGetRecord(db, index);
    recPtr = MemHandleLock(recHandle);
    DmWrite(recPtr, 0, ptr, size);
    MemPtrUnlock(recPtr); 
    DmReleaseRecord(db, index, true);  
}

static void readRecord(UInt16 index, UInt32 size, MemPtr ptr){
    MemPtr recPtr;
    MemHandle recHandle;

    recHandle = DmQueryRecord(db, index);
    recPtr = MemHandleLock(recHandle);
    MemMove(ptr, recPtr, size);
    MemPtrUnlock(recPtr); 
}
